# __init__.py -- Contrib module for Dulwich
# Copyright (C) 2014 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Contributed implementations and extensions for dulwich.

This package contains various contributed modules that extend dulwich
functionality. These modules are maintained as part of dulwich but may
have additional dependencies or more specialized use cases.

Available modules:
- greenthreads: Green-threaded support for finding missing objects
- paramiko_vendor: SSH client implementation using paramiko
- release_robot: Automated release management utilities
- requests_vendor: HTTP client implementation using requests
- swift: OpenStack Swift object storage backend
"""

__all__ = []
